from enum import Enum

from .grpc.search_pb2 import (
    CreateIndexedContentRequest,
    SearchIndexedContentRequest,
)
from .grpc.search_pb2_grpc import (
    CreateIndexedContentControllerStub,
    SearchIndexedContentControllerStub,
)

SUCCESS_CODE = "0"


def _singletonize_class_vars(cls):
    for attr_name, attr_value in cls.__dict__.items():
        if isinstance(attr_value, type):  # it's a class
            orig_cls = attr_value

            class SingletonWrapper(orig_cls):
                _instances = {}

                def __new__(cls, *args, **kwargs):
                    if cls not in cls._instances:
                        cls._instances[cls] = super().__new__(cls)
                    return cls._instances[cls]

            setattr(cls, attr_name, SingletonWrapper)
    return cls


class RequestMessages:
    CREATE = CreateIndexedContentRequest
    SEARCH = SearchIndexedContentRequest


class RPCMethodNames:
    CREATE = "Create"
    SEARCH = "Search"


@_singletonize_class_vars
class Stubs:
    CREATE = CreateIndexedContentControllerStub
    SEARCH = SearchIndexedContentControllerStub


class ItemTypes(str, Enum):
    BLOG = "blog"
    COURSE = "course"
    MASTER_COURSE = "master_course"
    LESSON = "lesson"



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: search/grpc/search.proto
# Protobuf Python Version: 4.25.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18search/grpc/search.proto\x12\x0e\x64\x61rsman.search\x1a\x1cgoogle/protobuf/struct.proto\"{\n\x1b\x43reateIndexedContentRequest\x12\x10\n\x08keywords\x18\x01 \x03(\t\x12\r\n\x05title\x18\x02 \x01(\t\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\x17\n\x0fitem_identifier\x18\x04 \x01(\t\x12\x11\n\titem_type\x18\x05 \x01(\t\"\x83\x01\n\x1c\x43reateIndexedContentResponse\x12\x14\n\x07message\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x11\n\x04\x63ode\x18\x02 \x01(\tH\x01\x88\x01\x01\x12%\n\x04\x64\x61ta\x18\x03 \x03(\x0b\x32\x17.google.protobuf.StructB\n\n\x08_messageB\x07\n\x05_code\"H\n\x1bSearchIndexedContentRequest\x12\x15\n\rsearch_phrase\x18\x01 \x01(\t\x12\x12\n\nitem_types\x18\x02 \x03(\t\"\x83\x01\n\x1cSearchIndexedContentResponse\x12\x14\n\x07message\x18\x01 \x01(\tH\x00\x88\x01\x01\x12\x11\n\x04\x63ode\x18\x02 \x01(\tH\x01\x88\x01\x01\x12%\n\x04\x64\x61ta\x18\x03 \x03(\x0b\x32\x17.google.protobuf.StructB\n\n\x08_messageB\x07\n\x05_code2\x87\x01\n\x1e\x43reateIndexedContentController\x12\x65\n\x06\x43reate\x12+.darsman.search.CreateIndexedContentRequest\x1a,.darsman.search.CreateIndexedContentResponse\"\x00\x32\x87\x01\n\x1eSearchIndexedContentController\x12\x65\n\x06Search\x12+.darsman.search.SearchIndexedContentRequest\x1a,.darsman.search.SearchIndexedContentResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'search.grpc.search_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_CREATEINDEXEDCONTENTREQUEST']._serialized_start=74
  _globals['_CREATEINDEXEDCONTENTREQUEST']._serialized_end=197
  _globals['_CREATEINDEXEDCONTENTRESPONSE']._serialized_start=200
  _globals['_CREATEINDEXEDCONTENTRESPONSE']._serialized_end=331
  _globals['_SEARCHINDEXEDCONTENTREQUEST']._serialized_start=333
  _globals['_SEARCHINDEXEDCONTENTREQUEST']._serialized_end=405
  _globals['_SEARCHINDEXEDCONTENTRESPONSE']._serialized_start=408
  _globals['_SEARCHINDEXEDCONTENTRESPONSE']._serialized_end=539
  _globals['_CREATEINDEXEDCONTENTCONTROLLER']._serialized_start=542
  _globals['_CREATEINDEXEDCONTENTCONTROLLER']._serialized_end=677
  _globals['_SEARCHINDEXEDCONTENTCONTROLLER']._serialized_start=680
  _globals['_SEARCHINDEXEDCONTENTCONTROLLER']._serialized_end=815
# @@protoc_insertion_point(module_scope)



from typing import ClassVar, List, Optional
from uuid import UUID

from google.protobuf.json_format import MessageToDict
from google.protobuf.message import Message
from pydantic import UUID4, BaseModel, validate_call

from . import config
from .api_constants import (
    SUCCESS_CODE,
    ItemTypes,
    RequestMessages,
    RPCMethodNames,
    Stubs,
)


class ValidationError(Exception):
    pass


class ResponseCodeError(Exception):
    pass


class RPCClient(BaseModel):
    _instances: ClassVar[dict] = {}
    stub: ClassVar

    def __new__(cls, *args, **kwargs):
        # Ensure that each subclass has its own singleton instance
        if cls not in cls._instances:
            cls._instances[cls] = super().__new__(cls)
        return cls._instances[cls]

    @validate_call(config=dict(arbitrary_types_allowed=True))
    async def perform_rpc(
        self,
        rpc_method_name: str,
        stub_cls,
        request_message: Optional[Message] = None,
        request_id: Optional[str | int | UUID] = None,
        timeout: Optional[float] = None,
        wait_for_ready: Optional[bool] = None,
    ):
        """Perform the remote-precedure-call.

        Parameters:
            rpc_method_name: The name of the RPC method (e.g., `Retrieve`).
            request_message: The request message instance.
            timeout: RPC timeout.
            wait_for_ready: Wait for Channel to be ready in case is not.
        """
        if config.CHANNEL is None:
            await config.construct_channel()

        RPCMethod = getattr(stub_cls(config.CHANNEL), rpc_method_name, None)

        if RPCMethod:
            metadata = [("api-key", config.grpc_settings.API_KEY)]
            if request_id:
                metadata.append(("x-request-id", str(request_id)))
            response = await RPCMethod(
                request_message,
                metadata=metadata,
                timeout=config.grpc_settings.TIMEOUT if timeout is None else timeout,
                wait_for_ready=config.grpc_settings.WAIT_FOR_READY
                if wait_for_ready is None
                else wait_for_ready,
            )

            # Raise error if response is not successful
            if response.code != SUCCESS_CODE:
                raise ResponseCodeError(response.code)

            return MessageToDict(response, preserving_proto_field_name=True)
        else:
            raise ValidationError("Invalid RPC method name.")


class IndexedContent(RPCClient):
    @validate_call
    async def create(
        self,
        title: str,
        content: str,
        item_identifier: str | UUID4,
        item_type: ItemTypes,
        keywords: Optional[List[str]] = None,
        request_id: Optional[str | int | UUID] = None,
        timeout: Optional[float] = None,
        wait_for_ready: Optional[bool] = None,
    ):
        """Create a new keyworded indexed content.
        Return:
            The search results in Darsman gRPC response format.
        """
        request = RequestMessages.CREATE(
            title=title,
            content=content,
            item_identifier=str(item_identifier),
            item_type=item_type.value,
        )
        if keywords:
            request.keywords.extend(keywords)

        return await self.perform_rpc(
            rpc_method_name=RPCMethodNames.CREATE,
            request_message=request,
            stub_cls=Stubs.CREATE,
            request_id=request_id,
            timeout=timeout,
            wait_for_ready=wait_for_ready,
        )

    async def create_blog(
        self,
        title: str,
        content: str,
        identifier: str | UUID4,
        keywords: Optional[List[str]] = None,
        request_id: Optional[str | int | UUID] = None,
        timeout: Optional[float] = None,
        wait_for_ready: Optional[bool] = None,
    ):
        return await self.create(
            title=title,
            content=content,
            item_identifier=identifier,
            item_type=ItemTypes.BLOG,
            keywords=keywords,
            request_id=request_id,
            timeout=timeout,
            wait_for_ready=wait_for_ready,
        )

    async def create_course(
        self,
        title: str,
        content: str,
        identifier: str | UUID4,
        keywords: Optional[List[str]] = None,
        request_id: Optional[str | int | UUID] = None,
        timeout: Optional[float] = None,
        wait_for_ready: Optional[bool] = None,
    ):
        return await self.create(
            title=title,
            content=content,
            item_identifier=identifier,
            item_type=ItemTypes.COURSE,
            keywords=keywords,
            request_id=request_id,
            timeout=timeout,
            wait_for_ready=wait_for_ready,
        )

    async def create_master_course(
        self,
        title: str,
        content: str,
        identifier: str | UUID4,
        keywords: Optional[List[str]] = None,
        request_id: Optional[str | int | UUID] = None,
        timeout: Optional[float] = None,
        wait_for_ready: Optional[bool] = None,
    ):
        return await self.create(
            title=title,
            content=content,
            item_identifier=identifier,
            item_type=ItemTypes.MASTER_COURSE,
            keywords=keywords,
            request_id=request_id,
            timeout=timeout,
            wait_for_ready=wait_for_ready,
        )

    async def create_lesson(
        self,
        title: str,
        content: str,
        identifier: str | UUID4,
        keywords: Optional[List[str]] = None,
        request_id: Optional[str | int | UUID] = None,
        timeout: Optional[float] = None,
        wait_for_ready: Optional[bool] = None,
    ):
        return await self.create(
            title=title,
            content=content,
            item_identifier=identifier,
            item_type=ItemTypes.LESSON,
            keywords=keywords,
            request_id=request_id,
            timeout=timeout,
            wait_for_ready=wait_for_ready,
        )

    @validate_call
    async def search(
        self,
        search_phrase: str,
        item_types: List[ItemTypes],
        request_id: Optional[str | int | UUID] = None,
        timeout: Optional[float] = None,
        wait_for_ready: Optional[bool] = None,
    ):
        """Search for a phrase across specified item types.

        Args:
            search_phrase (str): The phrase to search for.
            item_types (List[ItemTypes]): List of item types to filter the search.

        Return:
            The search results in Darsman gRPC response format.
        """
        request = RequestMessages.SEARCH(search_phrase=search_phrase)
        request.item_types.extend(item_types)

        return await self.perform_rpc(
            rpc_method_name=RPCMethodNames.SEARCH,
            request_message=request,
            stub_cls=Stubs.SEARCH,
            request_id=request_id,
            timeout=timeout,
            wait_for_ready=wait_for_ready,
        )

    async def search_blog(
        self,
        search_phrase: str,
        request_id: Optional[str | int | UUID] = None,
        timeout: Optional[float] = None,
        wait_for_ready: Optional[bool] = None,
    ):
        return await self.search(
            search_phrase=search_phrase,
            item_types=[ItemTypes.BLOG],
            request_id=request_id,
            timeout=timeout,
            wait_for_ready=wait_for_ready,
        )

    async def search_course_lesson(
        self,
        search_phrase: str,
        request_id: Optional[str | int | UUID] = None,
        timeout: Optional[float] = None,
        wait_for_ready: Optional[bool] = None,
    ):
        return await self.search(
            search_phrase=search_phrase,
            item_types=[
                ItemTypes.COURSE,
                ItemTypes.LESSON,
                ItemTypes.MASTER_COURSE,
            ],
            request_id=request_id,
            timeout=timeout,
            wait_for_ready=wait_for_ready,
        )
